/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.integrations.personInformation.impl;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationServiceException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class PersonInformationServiceClient {

	public static final String APPLICATION_JSON = "application/json";
	private static final String CONTENT_TYPE = "Content-Type";
	private static final int maximumUrlSize = 2048;
	private static final Logger LOGGER = Logger.getLogger(PersonInformationServiceClient.class);
	private final ObjectMapper om = new ObjectMapper();
	private final PersonInformationServiceClientFactory.ServiceErrorCallback serviceErrorCallback;
	private final HttpClientFactory httpClientFactory;
	private final String url;
	private final List<Integer> inactiveStatusList;
	private final ObjectMapper objectMapper = new ObjectMapper();
	protected enum requestTypes {STATUS, AGE, DECEASED};

	public PersonInformationServiceClient(HttpClientFactory httpClientFactory, String url, PersonInformationServiceClientFactory.ServiceErrorCallback serviceErrorCallback, List<Integer> inactiveStatusList) {
		this.url = url;
		this.httpClientFactory = httpClientFactory;
		this.serviceErrorCallback = serviceErrorCallback;
		this.inactiveStatusList = inactiveStatusList;
	}

	private HttpUriRequest buildDefaultRequest(String cpr, String requestType) {
		return new HttpGet(String.format("%s/person/%s/%s", url, cpr, requestType));
	}

	protected List<HttpUriRequest> buildDeceasedRequests(LinkedList<String> listOfCpr, Instant deceasedSince) {
		List<HttpUriRequest> requests = new ArrayList<>();
		String utcDateString = DateTimeFormatter.ISO_INSTANT.format(deceasedSince);
		int cprParamLength = "&cpr=".length();
		int estimatedSizePerCpr = 10 + cprParamLength; // Længden et CPR nummer fylder i et request
		int lengthOfUrlWithOutCpr = String.format("%s/person/deceased?&dato=%s", url, utcDateString).length(); // Længden af URL'en uden CPR numre
		int maxCprPerRequest = (maximumUrlSize - lengthOfUrlWithOutCpr) / estimatedSizePerCpr; // Giver antallet af CPR nummer som kan være i et request
		List<List<String>> groupsOfCpr = new ArrayList<>();

		// Opdel listOfCpr i grupper baseret på maxCprPerRequest
		for (int i = 0; i < listOfCpr.size(); i += maxCprPerRequest) {
			groupsOfCpr.add(listOfCpr.subList(i, Math.min(i + maxCprPerRequest, listOfCpr.size())));
		}

		// Opret en anmodning for hver gruppe
		for (List<String> group : groupsOfCpr) {
			String cprParams = group.stream().map(cpr -> "cpr=" + cpr).collect(Collectors.joining("&"));
			String fullUrl = String.format("%s/person/deceased?%s&dato=%s", url, cprParams, utcDateString);
			requests.add(new HttpGet(fullUrl));
		}

		return requests;
	}

	public PersonInformationCprStatus validateCpr(String patientId) {
		try {
			HttpUriRequest cprRequest = buildDefaultRequest(patientId, "status");
			PersonInformationCprStatus personInformationCprStatus = (PersonInformationCprStatus) invokeService(cprRequest, requestTypes.STATUS);
			serviceErrorCallback.registerServiceSucces();
			return personInformationCprStatus;
		} catch (ValidationServiceException e) {
			LOGGER.error("Fejl i kald til PersonInformation: ", e);
			serviceErrorCallback.registerServiceError();
			throw e;
		}
	}

	public int getAge(String patientId) {
		try {
			HttpUriRequest ageRequest = buildDefaultRequest(patientId, "age");
			int age = (int)invokeService(ageRequest, requestTypes.AGE);
			serviceErrorCallback.registerServiceSucces();
			return age;
		} catch (ValidationServiceException e) {
			LOGGER.error("Fejl i kald til PersonInformation: ", e);
			serviceErrorCallback.registerServiceError();
			throw e;
		}
	}

	public LinkedList<String> getDeceased(LinkedList<String> listOfCpr, Instant deceasedSince) {
		try {
			List<HttpUriRequest> deceasedRequests = buildDeceasedRequests(listOfCpr, deceasedSince);
			LinkedList<String> deceasedCprs = new LinkedList<>();
			for (HttpUriRequest request : deceasedRequests) {
				LinkedList<String> partialDeceased = (LinkedList<String>) invokeService(request, requestTypes.DECEASED);
				deceasedCprs.addAll(partialDeceased);
			}
			serviceErrorCallback.registerServiceSucces();
			return deceasedCprs;
		} catch (ValidationServiceException e) {
			LOGGER.error("Fejl i kald til PersonInformation: ", e);
			serviceErrorCallback.registerServiceError();
			throw e;
		}
	}

	protected Object invokeService(HttpUriRequest request, requestTypes requestType) {
		// Construct client
		HttpClient client = getClient();

		// Send request
		try {
			// Using a ResponseHandler makes the httpclient framework handle resource allocation/deallocation for us.
			// If you decide to not use a ResponseHandler anymore, you must ensure that resources (connections) are properly released!
			return client.execute(request, httpResponse -> handleResponse(httpResponse, requestType));
		}
		catch(IOException e) {
			throw new ValidationServiceException("CPR-nummeret kunne ikke valideres, fordi kaldet til PersonInformation-servicen gik galt.", e);
		}
	}

	private HttpClient getClient() {
		return httpClientFactory.getClient();
	}

	protected Object handleResponse(HttpResponse response, requestTypes requestType) {
		int statusCode = response.getStatusLine().getStatusCode();
		final String errorDescription = "CPR-nummeret eller alder kunne ikke valideres, fordi kaldet til " +
				"PersonInformation-servicen returnerede en uventet statuskode: %d";
		switch(statusCode) {
			case HttpStatus.SC_OK: {
				boolean contentTypeOk = contentTypeOk(response);
				if  (!contentTypeOk) {
					LOGGER.error("Fejl i retursvar fra PersonInformation (200)");
					throw new ValidationServiceException("Fejl i retursvar fra PersonInformation");
				}
				return parseResponseByType(response, requestType);
			}
			case HttpStatus.SC_NOT_FOUND: {
				boolean contentTypeOk = contentTypeOk(response);
				if  (!contentTypeOk) {
					LOGGER.error("Fejl i retursvar fra PersonInformation (404) forkert contenttype");
					throw new ValidationServiceException("Fejl i retursvar fra PersonInformation");
				}
				parseResponseAsJson(response);
				throw new ValidationServiceException(String.format(errorDescription, statusCode));
			}
			default:
				throw new ValidationServiceException(String.format(errorDescription, statusCode));
		}
	}

	private Object parseResponseByType(HttpResponse response, requestTypes requestType) {
		try {
			Map<String, Object> responseJson = null;
			switch (requestType) {
				case STATUS:
					responseJson = objectMapper.readValue(response.getEntity().getContent(), new TypeReference<Map<String, Object>>(){});
					return handleStatusResponse(responseJson);
				case AGE:
					responseJson = objectMapper.readValue(response.getEntity().getContent(), new TypeReference<Map<String, Object>>(){});
					return handleAgeResponse(responseJson);
				case DECEASED:
					return handleDeceasedResponse(response);
				default:
					LOGGER.error("Unhandled request type.");
					return null;
			}
		} catch (IOException e) {
			LOGGER.error("Error parsing response body to JSON", e);
			throw new ValidationServiceException("Error in interpreting response from PersonInformation", e);
		}
	}

	private Object handleStatusResponse(Map<String, Object> responseJson) {
		if (responseJson.get("status") instanceof Integer) {
			int status = (int) responseJson.get("status");
			LOGGER.debug("Status found for cpr is: " + status);
			return new PersonInformationCprStatus(inactiveStatusList.contains(status));
		}
		LOGGER.error("Status found for cpr is not a number.");
		return null;
	}

	private Object handleAgeResponse(Map<String, Object> responseJson) {
		if (responseJson.get("age") instanceof Integer) {
			int age = (int) responseJson.get("age");
			LOGGER.debug("Age found for cpr is: " + age);
			return age;
		}
		LOGGER.error("Age found for cpr is not a number.");
		return null;
	}

	private LinkedList<String> handleDeceasedResponse(HttpResponse response) throws IOException {
		List<Map<String, String>> list = objectMapper.readValue(response.getEntity().getContent(), new TypeReference<List<Map<String, String>>>(){});
		return list.stream().map(map -> map.get("cpr")).collect(Collectors.toCollection(LinkedList::new));
	}

	private Map<String, Object> parseResponseAsJson(HttpResponse response) {
		try {
			Map<String, Object> responseJson = om.readValue(response.getEntity().getContent(), Map.class);
			return responseJson;
		} catch (UnsupportedOperationException | IOException e) {
			LOGGER.error("Kunne ikke parse response body til json", e);
			throw new ValidationServiceException("Fejl i fortolkning af svar fra PersonInformation", e);
		}
	}

	protected boolean contentTypeOk(HttpResponse response) {
		Header headers = response.getFirstHeader(CONTENT_TYPE);
		boolean contentTypeOk = (headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON));
		return contentTypeOk;
	}
}
