/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.integrations.personInformation.impl;

import dk.sundhedsdatastyrelsen.livstestamente.util.HealthIndicator;
import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import org.apache.http.HttpStatus;

import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;

public class PersonInformationServiceClientFactory implements HealthIndicator {

	private String url;
	private HttpClientFactory httpClientFactory;
	private ServiceErrorCallback serviceErrorCallback;
	private int errorTolerance;
	private int statusErrorCode;
	private List<Integer> inactiveStatusList;

	public PersonInformationServiceClientFactory(String url, HttpClientFactory httpClientFactory, int statusErrorCode, List<Integer> inactiveStatusList) {
		this(url, httpClientFactory, 50, statusErrorCode, inactiveStatusList);
	}

	public PersonInformationServiceClientFactory(String url, HttpClientFactory httpClientFactory, int errorTolerance, int statusErrorCode, List<Integer> inactiveStatusList) {
		this.url = url;
		this.httpClientFactory = httpClientFactory;
		this.serviceErrorCallback = new ServiceErrorCallback();
		this.errorTolerance = errorTolerance;
		this.statusErrorCode = statusErrorCode;
		this.inactiveStatusList = inactiveStatusList;
	}

	public PersonInformationServiceClient createClient() {
		return new PersonInformationServiceClient(httpClientFactory, url, serviceErrorCallback, inactiveStatusList);
	}

	public static class ServiceErrorCallback {

		private ConcurrentLinkedQueue<LocalDateTime> errorTimeStamps = new ConcurrentLinkedQueue<LocalDateTime>();

		public void registerServiceError() {
			errorTimeStamps.add(LocalDateTime.now());				
		}

		public void registerServiceSucces() {
			errorTimeStamps.clear();
		}
	}

	private int countNumberOfErrors() {
		return serviceErrorCallback.errorTimeStamps.size();
	}

	@Override
	public boolean isHealthy() {
		// Fjern alle fejl, der er mere end 10 min gamle
		LocalDateTime tenMinutesAgo = LocalDateTime.now().minus(10, ChronoUnit.MINUTES);
		serviceErrorCallback.errorTimeStamps.removeIf(ldt -> ldt.isBefore(tenMinutesAgo));
		return countNumberOfErrors() < errorTolerance;
	}

	@Override
	public int getStatusCode() {
		return (isHealthy() ? HttpStatus.SC_OK : statusErrorCode);
	}

	@Override
	public String getName() {
		return "PersonInformationServiceClient";
	}

	@Override
	public String getStatusText() {
		return countNumberOfErrors()+" fejl";
	}
}