/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service;

import dk.sundhedsdatastyrelsen.livstestamente.util.HealthIndicator;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.LtrBtrContextListener;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;

import javax.servlet.ServletContext;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;

@WebServlet(name="IsAliveServlet", urlPatterns={"/isAlive"})
public class IsAliveServlet extends AutowiredHttpServlet {

    @SuppressWarnings("unused") private static final Logger logger = LogManager.getLogger(IsAliveServlet.class);

    @Autowired
    private DataSource btrDataSource;

    @Autowired(required = false)
    private Collection<HealthIndicator> healthIndicators = new ArrayList<>();

    private BuildData buildData;

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        int status;
        String statusMessage;
        String cr = System.getProperty("line.separator");

        StringBuilder b = new StringBuilder();
        if (datasourceResponding()) {
            status = 200;
            statusMessage = " OK";
            // Check other health indicators
            for (HealthIndicator healthIndicator : healthIndicators) {
                if(healthIndicator != null) {
                    int indicatorStatus = healthIndicator.getStatusCode();
                    if (indicatorStatus > status) {
                        status = indicatorStatus;
                        statusMessage = " Non-Authoritative Information from: " + healthIndicator.getName();
                    }
                }
            }
            b.append(status).append(statusMessage).append(cr);
        } else {
            status = 500;
            b.append(status).append(" Internal Server Error").append(cr);
            b.append("Database check failed").append(cr);
        }

        BuildData data = getBuildData(request.getServletContext());
        b.append(cr);
        b.append("Title: ").append(data.title).append(cr);
        b.append("Deployed: ").append(LtrBtrContextListener.DEPLOY_TIME).append(cr);
        b.append("Build-Date: ").append(buildData.date).append(cr);
        b.append("Build-Version: ").append(buildData.version).append(cr);
        b.append("Builder: ").append(buildData.builder).append(cr);
        b.append("Display time: ").append(Instant.now()).append(cr);
        b.append(getHealth());

        response.setStatus(status);
        response.setContentType("text/plain");
        response.getWriter().println(b);
    }

    private boolean datasourceResponding() {
        try {
            new JdbcTemplate(btrDataSource).execute("SELECT 1");
        } catch (DataAccessException e) {
            return false;
        }
        return true;
    }

    private BuildData getBuildData(ServletContext servletContext) throws IOException {
        if (buildData == null) {
            InputStream stream = servletContext.getResourceAsStream("/META-INF/MANIFEST.MF");
            if (stream == null) {
                // Local
                stream = this.getClass().getResourceAsStream("/META-INF/MANIFEST.MF");
            }
            if (stream != null) {
                Properties prop = new Properties();
                prop.load(new InputStreamReader(stream, "UTF-8"));
                stream.close();
                buildData = new BuildData(prop.getProperty("Implementation-Title"), prop.getProperty("Implementation-Version"), prop.getProperty("Built-Date"), prop.getProperty("Built-By"));
            }
        }
        return buildData;
    }

    private class BuildData {
        String title;
        String version;
        String date;
        String builder;

        BuildData(String title, String version, String date, String builder) {
            this.title = title;
            this.version = version;
            this.date = date;
            this.builder = builder;
        }
    }

    private String getHealth() {
        StringBuilder health = new StringBuilder();
        for (HealthIndicator healthIndicator : healthIndicators) {
            if(healthIndicator != null) {
                health.append(healthIndicator.getName()).append(" : \"").append(healthIndicator.getStatusText()).append("\"\n");
            }
        }
        return health.toString();
    }
}
